@@ -27,10 +27,6 @@
 #include <glib.h>
 #include <glib/gi18n.h>
 
-#ifdef HAVE_CLUTTER
-#include <clutter/clutter.h>
-#endif
-
 #include "ar-debug.h"
 #include "ar-runtime.h"
 #include "ar-string-utils.h"
@@ -74,7 +70,10 @@ struct _AisleriotGame
   GPtrArray *slots;
 
   char *game_file;
-  guint32 seed;
+
+  GRand *rand;
+  GRand *saved_rand;
+
   guint delayed_call_timeout_id;
 
   GTimer *timer;
@@ -243,16 +242,7 @@ clear_slots (AisleriotGame *game,
   for (i = 0; i < n_slots; ++i) {
     ArSlot *slot = game->slots->pdata[i];
 
-#ifdef HAVE_CLUTTER
-    if (slot->slot_renderer) {
-      clutter_actor_destroy (slot->slot_renderer);
-      g_object_unref (slot->slot_renderer);
-    }
-
-    g_byte_array_free (slot->old_cards, TRUE);
-#else
     g_ptr_array_free (slot->card_images, TRUE);
-#endif /* HAVE_CLUTTER */
     g_byte_array_free (slot->cards, TRUE);
 
     g_slice_free (ArSlot, slot);
@@ -307,7 +297,9 @@ cscmi_exception_get_backtrace (SCM tag, SCM throw_args)
   message = g_string_sized_new (1024);
 
   g_string_append_printf (message, "Variation: %s\n", aisleriot_game_get_game_file (game));
+#if 0
   g_string_append_printf (message, "Seed: %u\n", game->seed);
+#endif
 
   g_string_append (message, "Scheme error:\n\t");
 
@@ -660,11 +652,7 @@ cscmi_add_slot (SCM slot_data)
   slot->expanded_down = expanded_down != FALSE;
   slot->expanded_right = expanded_right != FALSE;
 
-#ifdef HAVE_CLUTTER
-  slot->old_cards = g_byte_array_sized_new (SLOT_CARDS_N_PREALLOC);
-#else
   slot->card_images = g_ptr_array_sized_new (SLOT_CARDS_N_PREALLOC);
-#endif
 
   slot->needs_update = TRUE;
 
@@ -679,7 +667,7 @@ static SCM
 scm_gettext (SCM message)
 {
   char *input, *output;
-  SCM translated;
+  SCM translated = SCM_UNDEFINED;
 
   if (!scm_is_string (message))
     return message;
@@ -925,7 +913,9 @@ scm_set_lambda (SCM start_game_lambda,
 static SCM
 scm_myrandom (SCM range)
 {
-  return scm_from_uint32 (g_random_int_range (0, scm_to_int (range)));
+  AisleriotGame *game = app_game;
+
+  return scm_from_uint32 (g_rand_int_range (game->rand, 0, scm_to_int (range)));
 }
 
 static SCM
@@ -1117,6 +1107,8 @@ cscmi_winning_game_lambda (void)
 static void
 aisleriot_game_init (AisleriotGame *game)
 {
+  game->rand = game->saved_rand = NULL;
+
   game->state = GAME_UNINITIALISED;
 
   game->slots = g_ptr_array_sized_new (SLOT_CARDS_N_PREALLOC);
@@ -1162,6 +1154,11 @@ aisleriot_game_finalize (GObject *object)
 
   g_timer_destroy (game->timer);
 
+  if (game->rand)
+    g_rand_free (game->rand);
+  if (game->saved_rand)
+    g_rand_free (game->saved_rand);
+
   app_game = NULL;
 
   G_OBJECT_CLASS (aisleriot_game_parent_class)->finalize (object);
@@ -1727,6 +1724,8 @@ game_scm_new_game (void *user_data)
   AisleriotGame *game = user_data;
   gboolean game_over;
 
+  g_assert (game->rand != NULL);
+
   /* It is possible for some games to not have any moves right from the
    * start. If this happens we redeal.
    */
@@ -1734,6 +1733,11 @@ game_scm_new_game (void *user_data)
   do {
     SCM size, lambda, over;
 
+    /* Copy RNG state */
+    if (game->saved_rand)
+      g_rand_free (game->saved_rand);
+    game->saved_rand = g_rand_copy (game->rand);
+
     size = scm_call_0 (game->lambdas[START_GAME_LAMBDA]);
     game->width = scm_to_double (SCM_CAR (size));
     game->height = scm_to_double (SCM_CADR (size));
@@ -1759,18 +1763,20 @@ game_scm_new_game (void *user_data)
   return SCM_BOOL_T;
 }
 
-/**
- * aisleriot_game_new_game:
+/*
+ * aisleriot_game_new_game_internal:
  * @game:
- * @seed: a pointer to a #guint, or %NULL
+ * @rand: (allow-none) (transfer full): a #GRand, or %NULL
+ * @update_statistics: whether to update statistic
  *
  * Starts a new game of the currently loaded game type.
- * If @seed is non-%NULL, the value it points to is used
- * as the game number; otherwise a random number is used.
+ *
+ * @game will take ownership of @rand.
  */
-void
-aisleriot_game_new_game (AisleriotGame *game,
-                         guint *seed)
+static void
+aisleriot_game_new_game_internal (AisleriotGame *game,
+                                  GRand *rand,
+                                  gboolean count_loss)
 {
   GObject *object = G_OBJECT (game);
   GError *err = NULL;
@@ -1791,17 +1797,24 @@ aisleriot_game_new_game (AisleriotGame *game,
   /* FIXMEchpe: this allows cheating the statistics by doing
    * Restart, then New Game.
    */
-  if (game->state >= GAME_RUNNING &&
-      (!seed || *seed != game->seed)) {
+  if (count_loss &&
+      game->state >= GAME_RUNNING) {
     update_statistics (game);
   }
 
-  if (seed) {
-    game->seed = *seed;
-  } else {
-    game->seed = 0;
+  if (rand != NULL) {
+    if (game->rand)
+      g_rand_free (game->rand);
+
+    game->rand = rand; /* adopted */
+  } else if (game->rand == NULL) {
+    game->rand = g_rand_new ();
   }
 
+  if (game->saved_rand)
+    g_rand_free (game->saved_rand);
+  game->saved_rand = NULL;
+
   clear_delayed_call (game);
   /* The game isn't actually in progress until the user makes a move */
   set_game_state (game, GAME_BEGIN);
@@ -1827,6 +1840,36 @@ aisleriot_game_new_game (AisleriotGame *game,
   g_signal_emit (game, signals[GAME_NEW], 0);
 }
 
+/**
+ * aisleriot_game_new_game:
+ * @game:
+ *
+ * Starts a new game of the currently loaded game type.
+ *
+ * @game will take ownership of @rand.
+ */
+void
+aisleriot_game_new_game (AisleriotGame *game)
+{
+  aisleriot_game_new_game_internal (game, NULL, TRUE);
+}
+
+/**
+ * aisleriot_game_new_game_with_rand:
+ * @game:
+ * @rand: (allow-none) (transfer full): a #GRand, or %NULL
+ *
+ * Starts a new game of the currently loaded game type.
+ *
+ * @game will take ownership of @rand.
+ */
+void
+aisleriot_game_new_game_with_rand (AisleriotGame *game,
+                                   GRand *rand)
+{
+  aisleriot_game_new_game_internal (game, rand, TRUE);
+}
+
 /**
  * aisleriot_game_restart_game:
  * @game:
@@ -1836,9 +1879,12 @@ aisleriot_game_new_game (AisleriotGame *game,
 void
 aisleriot_game_restart_game (AisleriotGame *game)
 {
-  guint seed = game->seed;
+  GRand *rand;
 
-  aisleriot_game_new_game (game, &seed);
+  rand = game->saved_rand;
+  game->saved_rand = NULL;
+
+  aisleriot_game_new_game_internal (game, rand, FALSE);
 }
 
 /**
@@ -1887,20 +1933,6 @@ aisleriot_game_get_geometry (AisleriotGame *game,
   *height = game->height;
 }
 
-/**
- * aisleriot_game_get_seed:
- * @game:
- *
- * Returns the seed of the current game.
- *
- * Returns: a number
- */
-guint
-aisleriot_game_get_seed (AisleriotGame *game)
-{
-  return game->seed;
-}
-
 /**
  * aisleriot_game_drag_valid:
  * @game:
@@ -2462,41 +2494,3 @@ aisleriot_game_deal_cards (AisleriotGame *game)
   aisleriot_game_end_move (game);
   aisleriot_game_test_end_of_game (game);
 }
-
-#ifdef HAVE_CLUTTER
-
-void
-aisleriot_game_get_card_offset (ArSlot *slot,
-                                guint card_num,
-                                gboolean old_cards,
-                                gint *xoff, gint *yoff)
-{
-  gint n_cards, exposed;
-
-  if (old_cards) {
-    n_cards = (gint) slot->old_cards->len;
-    exposed = (gint) slot->old_exposed;
-  } else {
-    n_cards = (gint) slot->cards->len;
-    exposed = (gint) slot->exposed;
-  }
-
-  if (card_num >= n_cards - exposed) {
-    gint idx = card_num + exposed - n_cards;
-    *xoff = slot->pixeldx * idx;
-    *yoff = slot->pixeldy * idx;
-  } else {
-    *xoff = 0;
-    *yoff = 0;
-  }
-}
-
-void
-aisleriot_game_reset_old_cards (ArSlot *slot)
-{
-  g_byte_array_set_size (slot->old_cards, 0);
-  g_byte_array_append (slot->old_cards, slot->cards->data, slot->cards->len);
-  slot->old_exposed = slot->exposed;
-}
-
-#endif /* HAVE_CLUTTER */
